#include <map>
#include <list>
#include <iostream>
#include <iomanip>

// Clase auxiliar Fecha
struct Fecha
{
    Fecha() {}
    Fecha(unsigned int dia, unsigned int mes, unsigned int anio) :
        dia(dia),
        mes(mes),
        anio(anio)
    {}

    bool operator==(const Fecha &f) const
    {
        return dia == f.dia && mes == f.mes && anio == f.anio;
    }

    bool operator<(const Fecha &f) const
    {
        return (anio < f.anio ||
                (anio == f.anio && mes < f.mes) ||
                (anio == f.anio && mes == f.mes && dia < f.dia));
    }

    unsigned int dia;
    unsigned int mes;
    unsigned int anio;
};

// Clase auxiliar Hora
struct Hora
{
    Hora() {}
    Hora(unsigned int hora, unsigned int minutos) :
        hora(hora),
        minutos(minutos)
    {}

    bool operator==(const Hora &h) const
    {
        return hora == h.hora && minutos == h.minutos;
    }

    bool operator<(const Hora &h) const
    {
        return (hora < h.hora || 
                (hora == h.hora && minutos < h.minutos));
    }

    unsigned int hora;
    unsigned int minutos;
};

std::ostream &operator<<(std::ostream &out, const Fecha &fecha)
{
    out << fecha.dia << "-" << fecha.mes << "-" << fecha.anio;
    return out;
}

std::ostream &operator<<(std::ostream &out,
                         const std::list<std::pair<Hora, std::string> > &list)
{
    for (std::list<std::pair<Hora, std::string> >::const_iterator i = 
             list.begin();
         i != list.end(); ++i) {
        out << std::setfill('0') << std::setw(2) << i->first.hora << ':' 
            << std::setw(2) << i->first.minutos << ": "
            << i->second << std::endl;
    }
    return out;
}

class Agenda
{
public:
    void insertarEvento(const Fecha &fecha, const Hora &hora, 
                        const std::string &descripcion);

    std::list<std::pair<Hora, std::string> > 
    eventosDia(const Fecha &fecha) const;

    std::list<std::pair<Hora, std::string> >
    eventosDia(const Fecha &fecha, const Hora & desde, const Hora &hasta) const;

    void eliminarEventos(const Fecha &desde_fecha, const Hora &desde_hora,
                         const Fecha &hasta_fecha, const Hora &hasta_hora);

protected:
    typedef std::map<Hora, std::string> EventosDia;
    typedef std::map<Fecha, EventosDia> Eventos;
    Eventos _eventos;
};

void Agenda::insertarEvento(const Fecha &fecha, const Hora &hora, 
                            const std::string &descripcion)
{
    // No comprobamos si había un evento en dicha fecha y hora. De ser
    // así, la nueva descripción reemplaza a la antigua.
    _eventos[fecha][hora] = descripcion;

}

std::list<std::pair<Hora, std::string> > 
Agenda::eventosDia(const Fecha &fecha) const
{
    std::list<std::pair<Hora, std::string> > lista;
    Eventos::const_iterator eventos = _eventos.find(fecha);
    if (eventos != _eventos.end()) {
        // Recorremos todos los eventos de la fecha indicada para añadirlos
        // a la lista. Si la fecha no estuviese en la agenda, entonces
        // no se haría nada.
        for (EventosDia::const_iterator e = eventos->second.begin();
             e != eventos->second.end(); ++e) {
            lista.push_back(*e);
        }
    }
    return lista;
}

std::list<std::pair<Hora, std::string> > 
Agenda::eventosDia(const Fecha &fecha,
                   const Hora &desde, const Hora &hasta) const
{
    std::list<std::pair<Hora, std::string> > lista;
    Eventos::const_iterator eventos = _eventos.find(fecha);
    if (eventos != _eventos.end()) {
        // Recorremos todos los eventos de la fecha indicada para añadirlos
        // a la lista. Si la fecha no estuviese en la agenda, entonces
        // no se haría nada.
        EventosDia::const_iterator inicio =
            eventos->second.lower_bound(desde);
        EventosDia::const_iterator fin =
            eventos->second.upper_bound(hasta);
        for (EventosDia::const_iterator e = inicio; e != fin; ++e)
            lista.push_back(*e);
    }
    return lista;
}


void Agenda::eliminarEventos(const Fecha &desde_fecha, const Hora &desde_hora,
                             const Fecha &hasta_fecha, const Hora &hasta_hora)
{
    Eventos::iterator inicio = _eventos.lower_bound(desde_fecha);
    Eventos::iterator fin = _eventos.upper_bound(hasta_fecha);
    
    if (inicio == _eventos.end())
        return;

    if (desde_fecha == hasta_fecha) {
        // Solo hay que borrar los eventos comprendidos entre dos horas.
        EventosDia &eventosDia = inicio->second;
        EventosDia::iterator desde = eventosDia.lower_bound(desde_hora);
        EventosDia::iterator hasta =  eventosDia.upper_bound(hasta_hora);
        eventosDia.erase(desde, hasta);
        // Si la collección de eventos para este día se queda vacía
        // entonces eliminamos el día de la agenda.
        if (eventosDia.empty())
            _eventos.erase(inicio);
    } else {
        // Hay que borrar los eventos comprendidos entre dos días distintos.

        // Primero eliminamos todos los eventos del día de inicio
        if (inicio->first == desde_fecha) {
            EventosDia &eventosDia = inicio->second;
            eventosDia.erase(eventosDia.lower_bound(desde_hora),
                             eventosDia.end());
        }
        
        // Ahora eliminamos todos los días que están completamente comprendidos
        // la fecha de inicio y la de fin.
        // fin apunta a un día después de la fecha dada (también puede ser 
        // end()) y como inicio != end entonces sabemos que --fin está definido.
        // Como queremos evaluar el último día aparte decrementamos fin y
        // comprobamos la fecha. Cuando la fecha es menor que la fecha final
        // quiere decir que la fecha final no está en la agenda, por lo que
        // volvemos a incrementar fin y no habrá que comprobar un día final.
        // Cuando fin es igual (no puede ser mayor) entonces borramos el 
        // intervalo [inicio + 1, fin).
        if ((--fin)->first < hasta_fecha)
            ++fin;
        _eventos.erase(++inicio, fin);

        if (fin != _eventos.end() && fin->first == hasta_fecha) {
            EventosDia &eventosDia = fin->second;
            eventosDia.erase(eventosDia.begin(),
                             eventosDia.upper_bound(hasta_hora));
        }
    }
}

int main(int argc, char *argv[])
{
    Agenda agenda;
    agenda.insertarEvento(Fecha(4, 10, 2009), Hora(12, 45), "reunión");
    agenda.insertarEvento(Fecha(4, 10, 2009), Hora(14, 20), "comida");
    agenda.insertarEvento(Fecha(4, 10, 2009), Hora(16, 20), "siesta");

    agenda.insertarEvento(Fecha(5, 10, 2009), Hora(8, 0), "conferencia");
    agenda.insertarEvento(Fecha(5, 10, 2009), Hora(15, 00), "2 pulgas y un café");
    agenda.insertarEvento(Fecha(5, 10, 2009), Hora(19, 00), "horas extra");
    agenda.insertarEvento(Fecha(5, 10, 2009), Hora(21, 00), "más horas extra");

    agenda.insertarEvento(Fecha(7, 10, 2009), Hora(14, 0), 
                          "comida con los suegros");

    agenda.insertarEvento(Fecha(8, 10, 2009), Hora(06, 30), "maratón");
    agenda.insertarEvento(Fecha(8, 10, 2009), Hora(13, 00), "fisioterapeuta");
    agenda.insertarEvento(Fecha(8, 10, 2009), Hora(15, 30), "traumatólogo");
    agenda.insertarEvento(Fecha(8, 10, 2009), Hora(17, 00), "baratón");


    std::cout << "Eventos del día " << Fecha(4, 10, 2009) << ": " << std::endl
              << agenda.eventosDia(Fecha(4, 10, 2009)) << std::endl;
    std::cout << "Eventos del día " << Fecha(8, 10, 2009) << ": " << std::endl
              << agenda.eventosDia(Fecha(8, 10, 2009)) << std::endl;

    // El día 5 parece duro
    std::cout << "Eventos de la mañana del día " << Fecha(5, 10, 2009) << ": "
              << std::endl
              << agenda.eventosDia(Fecha(5, 10, 2009),
                                   Hora(8, 0), Hora(14, 0));
    std::cout << "Eventos de la tarde-noche del día " << Fecha(5, 10, 2009) 
              << ": " << std::endl
              << agenda.eventosDia(Fecha(5, 10, 2009),
                                   Hora(15, 0), Hora(22, 0)) << std::endl;

    // Vamos a suavizarlo
    agenda.eliminarEventos(Fecha(5, 10, 2009), Hora(20, 00),
                           Fecha(5, 10, 2009), Hora(23, 00));
    std::cout << "Suavizado " << std::endl
              << agenda.eventosDia(Fecha(5, 10, 2009),
                                   Hora(0, 0), Hora(23, 59)) << std::endl;
    
    // Y el 8 es de locos
    // Vamos a quedarnos con lo que nos gusta
    agenda.eliminarEventos(Fecha(5, 10, 2009), Hora(9, 00),
                           Fecha(8, 10, 2009), Hora(16, 00));
    agenda.insertarEvento(Fecha(5, 10, 2009), Hora(18, 00),
                          "balneario");
    agenda.insertarEvento(Fecha(7, 10, 2009), Hora(14, 30),
                          "comilona de amigotes");

    std::cout << "Nos queda: " << std::endl
              << Fecha(5, 10, 2009) << ": " << std::endl
              << agenda.eventosDia(Fecha(5, 10, 2009))
              << Fecha(7, 10, 2009) << ": " << std::endl
              << agenda.eventosDia(Fecha(7, 10, 2009))
              << Fecha(8, 10, 2009) << ": " << std::endl
              << agenda.eventosDia(Fecha(8, 10, 2009));
}
